from __future__ import annotations

import abc
import contextlib
import inspect
import typing as t
from collections import defaultdict
from contextvars import ContextVar
from dataclasses import dataclass
from types import TracebackType

from ._contstants import DEFAULT_SCOPE
from ._exceptions import (
    InvalidMode,
    InvalidProviderType,
    InvalidScope,
    MissingProviderAnnotation,
    ProviderAlreadyBound,
    ScopeMismatch,
    UnknownProviderDependency,
)
from ._types import InterfaceT, Mode, Provider, Scope

ALLOWED_SCOPES: t.Dict[Scope, t.List[Scope]] = {
    "singleton": ["singleton"],
    "request": ["request", "singleton"],
    "transient": ["transient", "singleton", "request"],
}


class Dependency:
    __slots__ = ()


@dataclass(frozen=True)
class Binding:
    provider: Provider
    scope: Scope


@dataclass(frozen=True)
class LazyBinding:
    interface: t.Type[t.Any]
    parameter_name: str
    binding: Binding


class BaseDI(abc.ABC):
    mode: t.ClassVar[Mode]

    def __init__(self, default_scope: t.Optional[Scope] = None) -> None:
        self.default_scope = default_scope or DEFAULT_SCOPE
        self.bindings: t.Dict[t.Type[t.Any], Binding] = {}
        self.signature_cache: t.Dict[t.Callable[..., t.Any], inspect.Signature] = {}
        self.lazy_bindings: t.Dict[t.Type[t.Any], t.List[LazyBinding]] = defaultdict(
            list
        )

    def get_binding(self, interface: t.Type[InterfaceT]) -> Binding:
        try:
            return self.bindings[interface]
        except KeyError:
            raise LookupError(
                f"Binding to `{self.get_qualname(interface)}` "
                f"dependency is not registered."
            )

    def has_binding(self, interface: t.Type[InterfaceT]) -> bool:
        return interface in self.bindings

    def bind(
        self,
        interface: t.Type[InterfaceT],
        provider: Provider,
        *,
        scope: t.Optional[Scope] = None,
        override: bool = False,
    ) -> None:
        scope = scope or self.default_scope

        if not self.has_valid_scope(scope):
            raise InvalidScope(
                f"Invalid scope. Only {', '.join(t.get_args(Scope))} "
                "scope are supported."
            )

        if not self.has_valid_provider_type(provider):
            raise InvalidProviderType(
                "Invalid provider type. Only callable providers are allowed."
            )

        if self.mode == "sync" and (
            inspect.isasyncgenfunction(provider)
            or inspect.iscoroutinefunction(provider)
        ):
            raise InvalidMode(
                f"Cannot bind asynchronous provider "
                f"`{self.get_qualname(provider)}` in `sync` mode."
            )

        if self.has_binding(interface) and not override:
            raise ProviderAlreadyBound(
                f"Provider interface `{self.get_qualname(interface)}` already bound."
            )

        binding = Binding(provider=provider, scope=scope)

        self.validate_sub_providers(interface, binding)

        self.bindings[interface] = binding

    def provide(
        self, *, scope: t.Optional[Scope] = None, override: bool = False
    ) -> t.Callable[[Provider], t.Any]:
        def provider_func(provider: Provider) -> t.Any:
            interface = self.get_provider_annotation(provider)
            self.bind(interface, provider, scope=scope, override=override)

        return provider_func

    @staticmethod
    def has_valid_scope(scope: Scope) -> bool:
        return scope in t.get_args(Scope)

    @staticmethod
    def has_valid_provider_type(provider: Provider) -> bool:
        return callable(provider)

    def validate_sub_providers(
        self, interface: t.Type[InterfaceT], binding: Binding
    ) -> None:
        related_bindings = []

        provider, scope = binding.provider, binding.scope

        for parameter in self.get_signature(provider).parameters.values():
            if parameter.annotation is inspect._empty:  # noqa
                raise MissingProviderAnnotation(
                    f"Missing provider `{self.get_qualname(provider)}` "
                    f"dependency `{parameter.name}` annotation."
                )
            try:
                sub_binding = self.get_binding(parameter.annotation)
                related_bindings.append((sub_binding, True))
            except LookupError:
                self.lazy_bindings[parameter.annotation].append(
                    LazyBinding(
                        interface=interface,
                        parameter_name=parameter.name,
                        binding=binding,
                    )
                )

        for lazy_binding in self.lazy_bindings.pop(interface, []):
            try:
                sub_binding = self.get_binding(lazy_binding.interface)
                related_bindings.append((sub_binding, False))
            except LookupError:
                pass

        for related_binding, direct in related_bindings:
            if direct:
                left_scope, right_scope = related_binding.scope, scope
            else:
                left_scope, right_scope = scope, related_binding.scope
            allowed_scopes = ALLOWED_SCOPES.get(right_scope) or []
            if left_scope not in allowed_scopes:
                raise ScopeMismatch(
                    f"You tried to bind the `{scope}` scoped dependency "
                    f"`{self.get_qualname(provider)}` with "
                    f"a `{related_binding.scope}` scoped "
                    f"{self.get_qualname(related_binding.provider)}`."
                )

    def validate_bindings(self) -> None:
        if self.lazy_bindings:
            messages = []
            for lazy_interface, sub_bindings in self.lazy_bindings.items():
                for sub_binding in sub_bindings:
                    provider = sub_binding.binding.provider
                    parameter_name = sub_binding.parameter_name
                    provider_name = self.get_qualname(provider)
                    messages.append(
                        f"- `{provider_name}` has unknown `{parameter_name}`:"
                        f" `{self.get_qualname(lazy_interface)}` parameter"
                    )
            message = "\n".join(messages)
            raise UnknownProviderDependency(
                f"Unknown provider dependencies detected:\n{message}."
            )

    def get_provider_annotation(self, provider: Provider) -> t.Any:
        annotation = self.get_signature(provider).return_annotation

        if inspect.isclass(provider):
            return annotation

        if annotation is inspect._empty:  # noqa
            raise TypeError(
                f"Missing `{self.get_qualname(provider)}` provider return annotation."
            )

        origin = t.get_origin(annotation)
        args = t.get_args(annotation)

        # Supported generic types
        if origin in (list, dict, tuple):
            if args:
                return annotation
            else:
                raise TypeError(
                    f"Cannot use `{self.get_qualname(provider)}` generic type "
                    f"annotation without actual type."
                )

        try:
            return t.get_args(annotation)[0]
        except IndexError:
            return annotation

    def get_signature(self, obj: t.Callable[..., t.Any]) -> inspect.Signature:
        if (signature := self.signature_cache.get(obj)) is None:
            signature = inspect.signature(obj)
            self.signature_cache[obj] = signature
        return signature

    def get_injectable_params(self, obj: t.Callable[..., t.Any]) -> t.Dict[str, t.Any]:
        signature = self.get_signature(obj)
        parameters = signature.parameters
        params = {}
        for parameter in parameters.values():
            if not isinstance(parameter.default, Dependency):
                continue
            annotation = parameter.annotation
            if annotation is inspect._empty:  # noqa
                raise TypeError(
                    f"Missing `{self.get_qualname(obj)}` parameter annotation."
                )
            params[parameter.name] = annotation
        return params

    @staticmethod
    def get_qualname(obj: t.Any) -> str:
        qualname = obj.__qualname__
        module_name = getattr(obj, "__module__", "__main__")
        return f"{module_name}.{qualname}".removeprefix("builtins.")


class DI(BaseDI):
    mode = "sync"

    def __init__(self, default_scope: t.Optional[Scope] = None) -> None:
        super().__init__(default_scope)
        self.singleton_context = Context(self, scope="singleton")
        self.request_context_var: ContextVar[Context | None] = ContextVar(
            "request_context", default=None
        )

    def get(self, interface: t.Type[InterfaceT]) -> InterfaceT:
        binding = self.get_binding(interface)

        if binding.scope == "singleton":
            return self.singleton_context.get(interface)

        elif binding.scope == "request":
            request_context = self.request_context_var.get()
            if request_context is None:
                raise LookupError("Request context is not started.")
            return request_context.get(interface)

        # Transient scope
        return t.cast(InterfaceT, self.create_instance(interface))

    def close(self) -> None:
        self.singleton_context.close()

    @contextlib.contextmanager
    def request_context(self) -> t.Iterator[Context]:
        with Context(di=self, scope="request") as context:
            token = self.request_context_var.set(context)
            yield context
            self.request_context_var.reset(token)

    def inject_callable(self, obj: t.Callable[..., t.Any]) -> t.Callable[..., t.Any]:
        injectable_params = self.get_injectable_params(obj)

        def wrapped(*args: t.Any, **kwargs: t.Any) -> t.Any:
            for name, annotation in injectable_params.items():
                kwargs[name] = self.get(annotation)
            return obj(*args, **kwargs)

        return wrapped

    def create_instance(
        self,
        interface: t.Type[InterfaceT],
        stack: t.Optional[contextlib.ExitStack] = None,
    ) -> t.Any:
        provider = self.get_binding(interface).provider
        args, kwargs = self.get_provider_arguments(provider)
        if inspect.isgeneratorfunction(provider):
            cm = contextlib.contextmanager(provider)(*args, **kwargs)
            if stack:
                return stack.enter_context(cm)
            with contextlib.ExitStack() as stack:
                return stack.enter_context(cm)
        elif inspect.isfunction(provider):
            return provider(*args, **kwargs)
        return provider

    def get_provider_arguments(
        self, provider: Provider
    ) -> t.Tuple[t.List[t.Any], t.Dict[str, t.Any]]:
        args = []
        kwargs = {}
        signature = self.get_signature(provider)
        for parameter in signature.parameters.values():
            instance = self.get(parameter.annotation)
            if parameter.kind == parameter.POSITIONAL_ONLY:
                args.append(instance)
            else:
                kwargs[parameter.name] = instance
        return args, kwargs


class Context:
    def __init__(self, di: "DI", scope: Scope) -> None:
        self.di = di
        self.scope = scope
        self.instances: dict[t.Any, t.Any] = {}
        self.stack = contextlib.ExitStack()

    def get(self, interface: t.Type[InterfaceT]) -> InterfaceT:
        if (instance := self.instances.get(interface)) is None:
            instance = self.di.create_instance(interface, stack=self.stack)
            self.instances[interface] = instance
        return t.cast(InterfaceT, instance)

    def set(self, interface: t.Type[InterfaceT], instance: t.Any) -> None:
        self.di.bind(interface, instance, scope=self.scope)

    def close(self) -> None:
        self.stack.close()

    def __enter__(self) -> Context:
        return self

    def __exit__(
        self,
        exc_type: t.Type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        self.close()



import contextlib
import threading
import typing as t
from types import ModuleType

try:
    import anyio  # noqa

    anyio_installed = True
except ImportError:
    anyio_installed = False

from ._contstants import DEFAULT_MODE
from ._core import DI, Context, Dependency
from ._types import Mode, Provider, ProviderFunctionT, Scope
from ._utils import scan_package

if t.TYPE_CHECKING:
    from ._async import AsyncContext, AsyncDI


_di: t.Optional[DI] = None
_async_di: t.Optional["AsyncDI"] = None
_lock = threading.RLock()


dep = Dependency()


def _get_di() -> DI:
    global _di

    if _di is None:
        raise RuntimeError(
            '`pyxdi` not initialized. Please, call `pyxdi.init(mode="sync")` first.'
        )
    return _di


def _get_async_di() -> "AsyncDI":
    global _async_di

    if _async_di is None:
        raise RuntimeError(
            "`pyxdi` not initialized. " 'Please, call `pyxdi.init(mode="async")` first.'
        )
    return _async_di


def _get_di_or_async_di() -> t.Union[DI, "AsyncDI"]:
    global _di
    global _async_di

    if _di is None and _async_di is None:
        raise RuntimeError(
            "`pyxdi` not initialized. Please, call `pyxdi.init()` first."
        )
    return t.cast(t.Union[DI, "AsyncDI"], _di or _async_di)


def init(
    *,
    mode: t.Optional[Mode] = None,
    default_scope: t.Optional[Scope] = None,
    packages: t.Optional[t.Iterable[t.Union[ModuleType | str]]] = None,
    include: t.Optional[t.Iterable[str]] = None,
) -> None:
    global _di
    global _async_di

    if _di or _async_di:
        if _di:
            raise RuntimeError("`pyxdi` already initialized in `sync` mode.")
        elif _async_di:
            raise RuntimeError("`pyxdi` already initialized in `async` mode.")

    mode = mode or DEFAULT_MODE

    with _lock:
        if mode == "sync" and _di or mode == "async" and _async_di:
            return

        if mode == "async":
            if not anyio_installed:
                raise RuntimeError(
                    "Please, install `async` extension run in asynchronous mode. "
                    "eg. pip install pyxdi[async]"
                )

            from ._async import AsyncDI

            _async_di = AsyncDI(default_scope=default_scope)
        else:
            _di = DI(default_scope=default_scope)

        packages = packages or []
        for package in packages:
            scan_package(package, include=include)


def close() -> None:
    di = _get_di()
    return di.close()


async def aclose() -> None:
    di = _get_async_di()
    await di.close()


@contextlib.contextmanager
def request_context() -> t.Iterator[Context]:
    di = _get_di()
    with di.request_context() as ctx:
        yield ctx


@contextlib.asynccontextmanager
async def arequest_context() -> t.AsyncIterator["AsyncContext"]:
    di = _get_async_di()
    async with di.request_context() as ctx:
        yield ctx


@t.overload
def provider(
    func: None = ...,
    *,
    scope: Scope | None = None,
    override: bool = False,
) -> t.Callable[..., t.Any]:
    ...


@t.overload
def provider(
    func: ProviderFunctionT,
    *,
    scope: Scope | None = None,
    override: bool = False,
) -> t.Callable[[Provider], t.Any]:
    ...


def provider(
    func: t.Union[ProviderFunctionT, None] = None,
    *,
    scope: Scope | None = None,
    override: bool = False,
) -> t.Union[ProviderFunctionT, t.Callable[[Provider], t.Any]]:
    di = _get_di_or_async_di()
    provide = di.provide(scope=scope, override=override)

    if func is None:
        return provide
    elif callable(func):
        return provide(func)  # type: ignore[no-any-return]
    else:
        raise ValueError("Invalid provided dependency arguments.")


def inject(obj: t.Callable[..., t.Any]) -> t.Any:
    di = _get_di_or_async_di()
    return di.inject_callable(obj)



from ._api import (
    aclose,
    arequest_context,
    close,
    dep,
    init,
    inject,
    provider,
    request_context,
)

__all__ = [
    "aclose",
    "arequest_context",
    "close",
    "dep",
    "init",
    "inject",
    "provider",
    "request_context",
]



import importlib
import pkgutil
import typing as t
from types import ModuleType


def scan_package(
    package: t.Union[ModuleType, str], include: t.Optional[t.Iterable[str]] = None
) -> None:
    if isinstance(package, str):
        package = importlib.import_module(package)
    package_path = getattr(package, "__path__", None)
    if not package_path:
        return
    include = include or [""]
    for module_info in pkgutil.walk_packages(
        path=package_path, prefix=package.__name__ + "."
    ):
        for include_name in include:
            if module_info.name.endswith(include_name):
                importlib.import_module(module_info.name)



import typing as t

Provider = t.Callable[..., t.Any]
ProviderFunctionT = t.TypeVar("ProviderFunctionT", bound=t.Callable[..., t.Any])
Scope = t.Literal["transient", "singleton", "request"]
Mode = t.Literal["sync", "async"]
InterfaceT = t.TypeVar("InterfaceT")



class ProviderAlreadyBound(Exception):
    pass


class InvalidMode(ValueError):
    pass


class InvalidProviderType(Exception):
    pass


class InvalidScope(Exception):
    pass


class ScopeMismatch(Exception):
    pass


class MissingProviderAnnotation(Exception):
    pass


class UnknownProviderDependency(Exception):
    pass



from __future__ import annotations

import contextlib
import inspect
import typing as t
from contextvars import ContextVar
from functools import partial
from types import TracebackType

import anyio

from ._core import BaseDI
from ._types import InterfaceT, Provider, Scope


class AsyncDI(BaseDI):
    mode = "async"

    def __init__(self, default_scope: t.Optional[Scope] = None) -> None:
        super().__init__(default_scope)
        self.singleton_context = AsyncContext(self, scope="singleton")
        self.request_context_var: ContextVar[t.Optional[AsyncContext]] = ContextVar(
            "request_context", default=None
        )

    async def get(self, interface: t.Type[InterfaceT]) -> InterfaceT:
        binding = self.get_binding(interface)

        if binding.scope == "singleton":
            return await self.singleton_context.get(interface)

        elif binding.scope == "request":
            request_registry = self.request_context_var.get()
            if request_registry is None:
                raise LookupError("Request context is not started.")
            return await request_registry.get(interface)

        return t.cast(InterfaceT, await self.create_instance(interface))

    async def close(self) -> None:
        await self.singleton_context.close()

    @contextlib.asynccontextmanager
    async def request_context(self) -> t.AsyncIterator[AsyncContext]:
        async with AsyncContext(di=self, scope="request") as context:
            token = self.request_context_var.set(context)
            yield context
            self.request_context_var.reset(token)

    def inject_callable(self, target: t.Callable[..., t.Any]) -> t.Callable[..., t.Any]:
        injectable_params = self.get_injectable_params(target)

        async def wrapped(*args: t.Any, **kwargs: t.Any) -> t.Any:
            for name, annotation in injectable_params.items():
                kwargs[name] = await self.get(annotation)
            return await target(*args, **kwargs)

        def sync_wrapped(*args: t.Any, **kwargs: t.Any) -> t.Any:
            for name, annotation in injectable_params.items():
                kwargs[name] = anyio.from_thread.run(self.get, annotation)
            return target(*args, **kwargs)

        if inspect.iscoroutinefunction(target):
            return wrapped

        return sync_wrapped

    async def create_instance(
        self,
        interface: t.Type[InterfaceT],
        stack: contextlib.AsyncExitStack | None = None,
        sync_stack: contextlib.ExitStack | None = None,
    ) -> t.Any:
        dependency = self.get_binding(interface).provider
        args, kwargs = await self.get_provider_arguments(dependency)
        if inspect.isasyncgenfunction(dependency):
            acm = contextlib.asynccontextmanager(dependency)(*args, **kwargs)
            if stack:
                return await stack.enter_async_context(acm)
            async with contextlib.AsyncExitStack() as stack:
                return await stack.enter_async_context(acm)
        if inspect.isgeneratorfunction(dependency):
            cm = contextlib.contextmanager(dependency)(*args, **kwargs)
            if sync_stack:
                return await anyio.to_thread.run_sync(sync_stack.enter_context, cm)
            sync_stack = contextlib.ExitStack()
            try:
                return await anyio.to_thread.run_sync(sync_stack.enter_context, cm)
            finally:
                return await anyio.to_thread.run_sync(sync_stack.close)
        elif inspect.iscoroutinefunction(dependency):
            return await dependency(*args, **kwargs)
        elif inspect.isfunction(dependency):
            return await anyio.to_thread.run_sync(partial(dependency, *args, **kwargs))
        return dependency

    async def get_provider_arguments(
        self, provider: Provider
    ) -> tuple[list[t.Any], dict[str, t.Any]]:
        args = []
        kwargs = {}
        signature = inspect.signature(provider)
        for parameter in signature.parameters.values():
            instance = await self.get(parameter.annotation)
            if parameter.kind == parameter.POSITIONAL_ONLY:
                args.append(instance)
            else:
                kwargs[parameter.name] = instance
        return args, kwargs


class AsyncContext:
    def __init__(self, di: AsyncDI, scope: Scope) -> None:
        self.di = di
        self.scope = scope
        self.instances: dict[t.Any, t.Any] = {}
        self.stack = contextlib.AsyncExitStack()
        self.sync_stack = contextlib.ExitStack()

    async def get(self, interface: t.Type[InterfaceT]) -> InterfaceT:
        if (instance := self.instances.get(interface)) is None:
            instance = await self.di.create_instance(
                interface, stack=self.stack, sync_stack=self.sync_stack
            )
            self.instances[interface] = instance
        return t.cast(InterfaceT, instance)

    def set(self, interface: t.Type[InterfaceT], instance: t.Any) -> None:
        self.di.bind(interface, instance, scope=self.scope)

    async def close(self) -> None:
        await self.stack.aclose()
        await anyio.to_thread.run_sync(self.sync_stack.close)

    async def __aenter__(self) -> AsyncContext:
        return self

    async def __aexit__(
        self,
        exc_type: t.Type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        await self.close()



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': '.'}

packages = \
['pyxdi']

package_data = \
{'': ['*']}

install_requires = \
['anyio[asyncio]>=3.6.2,<4.0.0',
 'mkdocs-material[docs]>=9.0.12,<10.0.0',
 'mkdocs[docs]>=1.4.2,<2.0.0']

setup_kwargs = {
    'name': 'pyxdi',
    'version': '0.1.0a0',
    'description': 'Dependence Injection library',
    'long_description': '# Pyxdi (in dev)\n\n`Pyxdi` is a modern, lightweight and async-friendly Python Dependency Injection library that leverages type annotations ([PEP 484](https://peps.python.org/pep-0484/))\nto effortlessly manage dependencies in your applications, inspired by the functionality of [pytest fixtures](https://docs.pytest.org/en/7.2.x/explanation/fixtures.html).\n\n[![CI](https://github.com/antonrh/pyxdi/actions/workflows/ci.yml/badge.svg)](https://github.com/antonrh/pyxdi/actions/workflows/ci.yml)\n[![codecov](https://codecov.io/gh/antonrh/pyxdi/branch/main/graph/badge.svg?token=67CLD19I0C)](https://codecov.io/gh/antonrh/pyxdi)\n[![Documentation Status](https://readthedocs.org/projects/pyxdi/badge/?version=latest)](https://pyxdi.readthedocs.io/en/latest/?badge=latest)\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n\n---\nDocumentation\n\nhttp://pyxdi.readthedocs.io/\n\n---\n\n## Installing\n\nInstall using `pip`:\n\n```bash\npip install pyxdi\n```\n\nor using `poetry`:\n\n```bash\npoetry add pyxdi\n```\n\n## Quick Example\n\n*app.py*\n\n```python\nimport pyxdi\n\npyxdi.init()\n\n\n@pyxdi.provider\ndef message() -> str:\n    return "Hello, world!"\n\n\n@pyxdi.inject\ndef say_hello(message: str = pyxdi.dep) -> None:\n    print(message)\n\n\nif __name__ == "__main__":\n    say_hello()\n```\n\n---\n\n## TODO\n* Unit tests\n* Autowiring\n* Documentation\n* Examples\n* Pypi package\n',
    'author': 'Anton Ruhlov',
    'author_email': 'antonruhlov@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/antonrh/pyxdi',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<3.12',
}


setup(**setup_kwargs)
